import openai
from abstract_security.envy_it import get_env_value
class ApiManager:
    """
    This class handles the management of OpenAI API keys and headers.
    It is capable of retrieving the API key from environment variables, loading it for use in requests,
    and constructing authorization headers for the API calls.

    Attributes:
        content_type (str): The type of the content being sent in the request. Defaults to 'application/json'.
        api_env (str): The name of the environment variable containing the API key. Defaults to 'OPENAI_API_KEY'.
        api_key (str): The API key for OpenAI. If not provided, it is retrieved using the `get_openai_key` method.
        header (dict): The headers to be used in the requests. If not provided, they are generated using the 'get_header' method.

    Methods:
        get_openai_key(): Retrieves OpenAI API key from environment variables.
        load_openai_key(): Loads the OpenAI API key for use in requests.
        get_header(): Generates the request headers for the API calls.
    """
    def __init__(self,api_key:str=None,api_env:str=None,content_type=None,header=None,env_path=None)->None:
        self.content_type=content_type or 'application/json'
        self.api_env=api_env or 'OPENAI_API_KEY'
        self.env_path = env_path
        self.api_key=api_key or self.get_openai_key()
        self.header=header or self.get_header()
        self.load_openai_key()
    def get_openai_key(self)->str:
        """
        Retrieves the OpenAI API key from the environment variables.

        Args:
            key (str): The name of the environment variable containing the API key. 
                Defaults to 'OPENAI_API_KEY'.

        Returns:
            str: The OpenAI API key.
        """
        return get_env_value(key=self.api_env,path=self.env_path)
    def load_openai_key(self)->None:
        """
        Loads the OpenAI API key for authentication.
        """
        openai.api_key = self.api_key
    def get_header(self)->dict:
        """
        Generates request headers for API call.
        
        Args:
            content_type (str): Type of the content being sent in the request. Default is 'application/json'.
            api_key (str): The API key for authorization. By default, it retrieves the OpenAI API key.
            
        Returns:
            dict: Dictionary containing the 'Content-Type' and 'Authorization' headers.
        """
        return {'Content-Type': self.content_type, 'Authorization': f'Bearer {self.api_key}'}
    



from abstract_utilities import (os,
                                get_date,
                                mkdirs,
                                safe_json_loads,
                                safe_read_from_json)
import json

class SaveManager:
    """
    Manages the saving of data. This class should provide methods to specify where (e.g., what database or file) and how (e.g., in what format) data should be saved.
    """
    def __init__(self, data={},title:str=None,directory:str=None,model:str='default')->None:
        self.title=title
        self.model=model
        self.date = get_date()
        if isinstance(directory,list):
            abs_path = directory[0]
            if not os.path.isabs(directory[0]):
                abs_path = os.path.join(os.getcwd(), directory[0])
                mkdirs(abs_path)
            path = abs_path
            for child in directory[1:]:
                path = os.path.join(path, child)
                mkdirs(path)
            self.directory=path
        else:
            self.directory = mkdirs(directory or os.path.join(os.getcwd(), 'response_data'))
            self.directory = mkdirs(os.path.join(self.directory, self.date))
            self.directory = mkdirs(os.path.join(self.directory, self.model))
        self.file_name = self.create_unique_file_name()
        self.file_path = os.path.join(self.directory, self.file_name)
        if data:
            self.data = safe_json_loads(data)
            self.data['file_path']=self.file_path
            self.data['title']=self.title
            self.data['model']=self.model
            self.save_to_file(data = data,file_path = self.file_path)
    def create_unique_file_name(self) -> str:
        # Sanitize and shorten the title
        sanitized_title = self.sanitize_title(self.title)

        # Generate base file name
        base_name = f"{sanitized_title}.json"
        
        # Check for uniqueness and append index if needed
        unique_name = base_name
        index = 1
        while os.path.exists(os.path.join(self.directory, unique_name)):
            unique_name = f"{sanitized_title}_{index}.json"
            index += 1

        return unique_name
    @staticmethod
    def sanitize_title(title: str) -> str:
        if title:
            # Replace spaces and special characters
            title = str(title).replace(" ", "_").replace(":", "_")

            # Limit the length of the title
            max_length = 30
            if len(title) > max_length:
                title = title[:max_length]

            return title
    def save_to_file(self, data:dict, file_path:str)->None:
        # Assuming `data` is already a dictionary, we convert it to a JSON string and save.
        with open(file_path, 'w', encoding='utf-8') as file:
            json.dump(data, file, ensure_ascii=False, indent=4)
    
    @staticmethod
    def read_saved_json(file_path:str)->dict:
        # Use 'safe_read_from_json' which is presumed to handle errors and return JSON
        return safe_read_from_json(file_path)
    
    @staticmethod
    def read_saved_json(file_path:str)->dict:
        # Use 'safe_read_from_json' which is presumed to handle errors and return JSON
        return safe_read_from_json(file_path)



from abstract_utilities import get_actual_number, eatAll
from abstract_utilities.type_utils import is_bool
import json

def get_dict_from_instruction(instructions):
    new_instructions = {}
    if instructions:
        count = 0
        if isinstance(instructions, str):
            new_instructions[f"additional_instructions_{count}"] = {
                "instruction": instructions,
                "example": "",
                "default": True
            }
        elif isinstance(instructions, list):
            for instruction in instructions:
                if isinstance(instruction, str):
                    new_instructions[f"additional_instructions_{count}"] = {
                        "instruction": instruction,
                        "example": "",
                        "default": True
                    }
                    count += 1
                if isinstance(instruction, dict):
                    new_instruction = {"instruction": "", "example": "", "default": True}
                    new_instruction.update(instruction)
                    new_instructions[f"additional_instructions_{count}"] = new_instruction
                    count += 1
        elif isinstance(instructions, dict):
            for key, value in instructions.items():
                new_instruction = {"instruction": "", "example": "", "default": True}
                new_instruction.update(instructions[key])
                new_instructions[key] = new_instruction
    return new_instructions

class InstructionManager:
    def __init__(self, instructions=None, instruction_bools=None) -> None:
        instruction_bools = instruction_bools or {}
        self.default_instructions = {
            "api_response": {
                "instruction": "Place response to prompt here.",
                "example": "This is the reply to your request.",
                "default": True
            },
            "additional_responses": {
                "instruction": "Marking 'True' initiates a loop which continues to send the current chunk's prompt until the module returns a 'False' value.",
                "example": False,  # Use Python boolean for consistency
                "default": False
            },
            "generate_title": {
                "instruction": "Generate a short and concise title for this query, using underscores instead of spaces and avoiding special characters. The title should not exceed 200 characters to ensure compatibility with file systems.",
                "example": "Short_Title_for_Query",
                "default": True
            },
            "database_query": {
                "instruction": "The given database schema, variables, values, and function are provided in the prompt data; please derive the most appropriate inputs for a database query from the context and assertions of the request. Place all inputs for the function within this response key.",
                "example": {"query": """SELECT dnc_data FROM dncdata WHERE dnc_data ->> 'Data_Subscriber_First_Name' = 'Karen';"""},
                "default": False
            },
            "notation": {
                "instruction": "Notation is a module-end functionality that allows the module to preserve relevant information or context for subsequent prompts, enabling communication between modules throughout the query iterations.",
                "example": "Selecting additional responses due to insufficient completion tokens.",
                "default": False
            },
            "suggestions": {
                "instruction": "This parameter allows the module to provide suggestions for improving efficiency in future prompt sequences.",
                "example": "Consider batching queries to reduce server load",
                "default": False
            },
            "abort": {
                "instruction": "If you cannot fulfill the request, respond with this value as 'True'. Leave a detailed reason as to why the query stream was aborted in 'suggestions'.",
                "example": False,  # Use Python boolean
                "default": False
            },
            "prompt_as_previous": {
                "instruction": "This is a user-end declaration. If this is visible, the request portion of the prompt will change to include previous response data, if needed.",
                "example": False,  # Use Python boolean
                "default": False
            },
            "request_chunks": {
                "instruction": "Request to prompt again the previous chunk data. If selected, the query will iterate once more with the previous data chunk included in the prompt.",
                "example": False,  # Use Python boolean
                "default": False
            },
            "token_adjustment": {
                "instruction": "Suggest percentage adjustments, between -100 up to 100, for the future token allotment. If it will provide better results to increase or decrease the future allotment, place a number.",
                "example": 0,
                "default": False
            },
        }
        self.instructions = [[]]
        self.default_instructions.update(get_dict_from_instruction(instructions))
        for key, value in instruction_bools.items():
            if key in self.default_instructions:
                self.default_instructions[key]["default"] = value
        self.update_instructions(0, **{
            key: self.default_instructions[key]
            for key in self.default_instructions
            if self.default_instructions[key].get("default")
        })

    def get_instructions(self, instruction_number: int = None) -> dict:
        if not self.instructions or self.instructions == [[]]:
            self.add_instructions()
        instruction_number = instruction_number if instruction_number is not None else -1
        return self.instructions[instruction_number]

    def update_default_instructions(self, data):
        self.default_instructions.update(get_dict_from_instruction(data))
        self.update_instructions(0, **{
            key: self.default_instructions[key]
            for key in self.default_instructions
            if self.default_instructions[key].get("default")
        })

    def get_instructions_bools(self, parameters: dict) -> dict:
        return {parameter: bool(value) for parameter, value in parameters.items()}

    def get_instructions_text_values(self, instruction_bools: dict, parameters: dict) -> dict:
        parameter_text_values = {}
        for parameter, value in instruction_bools.items():
            if value and parameter in self.default_instructions:
                parameter_text_values[parameter] = (
                    self.default_instructions[parameter]["instruction"]
                    if is_bool(parameters[parameter])
                    else parameters[parameter]
                )
        return parameter_text_values

    def get_instructions_text(self, instructions_js: dict) -> tuple[str, dict]:
        """
        Retrieves instructions and example dictionary for the conversation.

        Args:
            instructions_js: Dictionary of instruction text values.

        Returns:
            Tuple of (instructions string, example dictionary).
        """
        instructions = ""
        example_format = {}
        if instructions_js:
            instructions = (
                "Your response is expected to be in JSON format, with boolean responses as lowercase unquoted values (true/false), "
                "using the following keys:\n\n"
            )
            for i, key in enumerate(instructions_js.keys()):
                if key in self.default_instructions:
                    instructions += f'{i}) {key} - """{eatAll(instructions_js[key], ["\'"])}"""\n'
                    example_value = self.default_instructions[key]["example"]
                    # Ensure JSON-compliant booleans
                    if isinstance(example_value, bool):
                        example_value = str(example_value).lower()
                    example_format[key] = example_value
            instructions += (
                "\nBelow is an example of the expected JSON dictionary response format, with default inputs. "
                "This response MUST be a dict:\n" + json.dumps(example_format, indent=2)
            )
        return instructions, example_format

    def add_instructions(self, all_true: bool = False, **kwargs) -> None:
        if not self.instructions or self.instructions == [[]]:
            instruction_display, instruction_bools, instructions_text_values, instructions_text, example_format = (
                self.get_instructions_values(all_true, **kwargs)
            )
            last_instructions = {
                "instruction_display": instruction_display,
                "instructions_bools": instruction_bools,
                "instructions_text_values": instructions_text_values,
                "instructions_text": instructions_text,
                "example_format": example_format,
                "text": instructions_text
            }
        else:
            last_instructions = self.instructions[-1]
        self.instructions.append(last_instructions)

    def get_instructions_values(self, all_true: bool = False, **kwargs) -> tuple[bool, dict, dict, str, dict]:
        parameters = {}
        for parameter in self.default_instructions.keys():
            parameters[parameter] = kwargs.get(parameter, False) if not all_true else True
        instruction_display = kwargs.get("instruction_display", True)
        instruction_bools = self.get_instructions_bools(parameters)
        instructions_text_values = self.get_instructions_text_values(instruction_bools, parameters)
        if not instructions_text_values:
            # Ensure at least one instruction is included to avoid empty prompt
            instructions_text_values["api_response"] = self.default_instructions["api_response"]["instruction"]
            instruction_bools["api_response"] = True
        instructions_text, example_format = self.get_instructions_text(instructions_js=instructions_text_values)
        return instruction_display, instruction_bools, instructions_text_values, instructions_text, example_format

    def update_instructions(self, instruction_number: int = 0, **kwargs) -> None:
        instruction_display, instruction_bools, instructions_text_values, instructions_text, example_format = (
            self.get_instructions_values(**kwargs)
        )
        while len(self.instructions) <= instruction_number:
            self.instructions.append({})
        self.instructions[instruction_number] = {
            "instruction_display": instruction_display,
            "instructions_bools": instruction_bools,
            "instructions_text_values": instructions_text_values,
            "instructions_text": instructions_text,
            "example_format": example_format,
            "text": instructions_text
        }

    def parse_model_response(self, response: str) -> dict:
        """
        Parses a model response string into a dictionary.

        Args:
            response: The model's response string (expected to be JSON-formatted).

        Returns:
            Dictionary containing the parsed response, or an error message if parsing fails.
        """
        try:
            return json.loads(response)
        except json.JSONDecodeError as e:
            return {
                "error": f"Failed to parse JSON response: {str(e)}",
                "raw_response": response
            }



import os
import json
import PySimpleGUI as psg
import execjs  # For running JavaScript in Python

# JavaScript code to parse JS/TS/JSX/TSX files using @babel/parser
JS_CODE = """
const parser = require('@babel/parser');
const traverse = require('@babel/traverse').default;
const types = require('@babel/types');

function parseCode(code) {
  const ast = parser.parse(code, {
    sourceType: 'module',
    plugins: ['jsx', 'typescript'],
  });

  const imports = [];
  const components = [];
  const dependencies = {};

  traverse(ast, {
    // Extract import declarations
    ImportDeclaration(path) {
      const source = path.node.source.value;
      path.node.specifiers.forEach(spec => {
        imports.push({
          name: spec.local.name,
          source: source,
          type: spec.type // e.g., ImportSpecifier, ImportDefaultSpecifier
        });
      });
    },

    // Extract React functional components (functions returning JSX)
    FunctionDeclaration(path) {
      const name = path.node.id.name;
      const hasJSX = path.node.body.body.some(statement => {
        return types.isReturnStatement(statement) && 
               statement.argument && 
               types.isJSXElement(statement.argument);
      });
      if (hasJSX) {
        components.push({ name, type: 'functional' });
        dependencies[name] = collectDependencies(path.node.body);
      }
    },

    // Extract React class components
    ClassDeclaration(path) {
      const name = path.node.id.name;
      const isComponent = path.node.superClass && (
        path.node.superClass.name === 'Component' ||
        path.node.superClass.name === 'PureComponent' ||
        (path.node.superClass.object && path.node.superClass.object.name === 'React' && 
         path.node.superClass.property.name === 'Component')
      );
      if (isComponent) {
        components.push({ name, type: 'class' });
        dependencies[name] = collectDependencies(path.node.body);
      }
    },
  });

  function collectDependencies(node) {
    const deps = new Set();
    traverse(node, {
      Identifier(path) {
        // Collect identifiers that might be dependencies (e.g., other components, hooks)
        if (!path.parentPath.isVariableDeclarator() && !path.parentPath.isFunctionDeclaration()) {
          deps.add(path.node.name);
        }
      },
      MemberExpression(path) {
        // Collect member expressions (e.g., React.useState)
        if (path.node.object.name === 'React' || path.node.object.name === 'use') {
          deps.add(`${path.node.object.name}.${path.node.property.name}`);
        }
      },
    }, { scope: null });
    return Array.from(deps);
  }

  return { imports, components, dependencies };
}
"""

# Compile the JavaScript code
ctx = execjs.compile(JS_CODE)

def parse_react_file(filepath):
    """Parse a JS/TS/JSX/TSX file and return imports, components, and dependencies."""
    with open(filepath, 'r', encoding='utf-8') as f:
        code = f.read()
    
    try:
        result = ctx.call('parseCode', code)
        return result['imports'], result['components'], result['dependencies']
    except Exception as e:
        print(f"Error parsing file {filepath}: {e}")
        return [], [], {}

def get_import_lines(filepath):
    """Extract import statements from the file."""
    with open(filepath, 'r', encoding='utf-8') as f:
        lines = f.readlines()
    return [line.strip() for line in lines if line.strip().startswith(('import ', 'from '))]

def get_component_code(filepath, component_name):
    """Extract the code block for a specific component."""
    with open(filepath, 'r', encoding='utf-8') as f:
        lines = f.readlines()

    code = ''
    recording = False
    indentation = ''
    brace_count = 0

    for line in lines:
        stripped = line.lstrip()
        # Start recording for function or class component
        if stripped.startswith(f'function {component_name}') or \
           stripped.startswith(f'const {component_name} =') or \
           stripped.startswith(f'class {component_name}'):
            recording = True
            indentation = line[:len(line) - len(stripped)]
            code += line
            if '{' in line:
                brace_count += line.count('{') - line.count('}')
        elif recording:
            code += line
            brace_count += (line.count('{') - line.count('}'))
            # Stop recording when the component block ends
            if brace_count == 0 and line.strip() == '':
                break
        # Stop if another function/class is encountered
        elif line.startswith(indentation + 'function ') or \
             line.startswith(indentation + 'const ') or \
             line.startswith(indentation + 'class '):
            if recording:
                break

    return code

class ReactAnalyser:
    def __init__(self, filepath):
        self.filepath = filepath
        self.imports, self.components, self.dependencies = parse_react_file(filepath)

    def get_components(self):
        return self.components

    def get_imports(self):
        return self.imports

    def get_component_deps(self, component_name):
        """Return dependencies for a specific component."""
        deps = self.dependencies.get(component_name, [])
        # Filter dependencies to include only known components or imports
        all_component_names = [c['name'] for c in self.components]
        all_import_names = [i['name'] for i in self.imports]
        return [dep for dep in deps if dep in all_component_names or dep in all_import_names]

# GUI Layout
layout = [
    [psg.FileBrowse(file_types=(("JavaScript/TypeScript Files", "*.js *.jsx *.ts *.tsx"),), key='-FILEBROWSE-')],
    [psg.Button("Load Components")],
    [psg.Listbox(values=[], size=(40, 10), key='-LIST-', enable_events=True, select_mode='multiple')],
    [psg.Multiline(size=(60, 20), key='-MULTI-', disabled=True)],
    [psg.Text('Save Location'), psg.InputText(os.getcwd(), key='-SAVEFOLDER-', disabled=True), psg.FolderBrowse('Browse', key='-FOLDERBROWSE-')],
    [psg.Text('Output Filename'), psg.InputText("output", key='-OUTPUTFILE-')],
    [psg.Button('Make Components'), psg.Button('Grab Dependencies'), psg.Button('Exit')],
]

window = psg.Window('React Component Browser', layout)

while True:
    event, values = window.read()
    
    if event in (psg.WINDOW_CLOSED, 'Exit'):
        break

    if event == 'Load Components':
        file = values['-FILEBROWSE-']
        if file.endswith(('.js', '.jsx', '.ts', '.tsx')):
            analyser = ReactAnalyser(file)
            window['-LIST-'].update([comp['name'] for comp in analyser.get_components()])
        else:
            psg.popup("Please select a JavaScript/TypeScript file.")

    if event == '-FOLDERBROWSE-':
        folder = values['-FOLDERBROWSE-']
        if folder:
            window['-SAVEFOLDER-'].update(folder)
            output_filename = values['-OUTPUTFILE-']
            if output_filename == 'output':
                i = 1
                while os.path.isfile(os.path.join(folder, f'{output_filename}_{i}.tsx')):
                    i += 1
                output_filename = f'{output_filename}_{i}'
            window['-OUTPUTFILE-'].update(output_filename)

    if event == '-LIST-':
        selected_components = values['-LIST-']
        analyser = ReactAnalyser(file)
        all_imports = set()
        all_deps = set()
        for comp in selected_components:
            deps = analyser.get_component_deps(comp)
            all_deps.update(deps)
            all_imports.update([i['name'] for i in analyser.get_imports() if i['name'] in deps])
        all_deps = list(all_imports) + list(all_deps)
        for comp in selected_components:
            if comp not in all_deps:
                all_deps.append(comp)
        window['-MULTI-'].update("\n".join(sorted(all_deps)))

    if event == 'Make Components':
        folder = values['-SAVEFOLDER-']
        output_filename = values['-OUTPUTFILE-']
        output_file = os.path.join(folder, f'{output_filename}.tsx')

        names = window['-MULTI-'].get().split("\n")
        import_lines = get_import_lines(file)

        with open(output_file, 'w', encoding='utf-8') as f:
            f.writelines(line + '\n' for line in import_lines)
            f.write("\n")
            for name in names:
                comp_code = get_component_code(file, name)
                if comp_code:
                    f.write(comp_code)
                    f.write("\n\n")

        psg.popup(f"Components saved to {output_file}.")

window.close()
